import argparse
import os
import sys
from typing import Tuple

import sambaflow.samba.utils as utils
import torch
import torch.nn as nn
import torchvision
import torchvision.transforms as transforms
from sambaflow import samba
from sambaflow.models.torch_models.logreg_torch import LogRegTorch
from sambaflow.samba.utils.argparser import parse_app_args
from sambaflow.samba.utils.common import common_app_driver
from sambaflow.samba.utils.dataset.mnist import dataset_transform


class ConvNet(nn.Module):
    def __init__(self, num_classes=10, input_shape=[28, 28]):
        super(ConvNet, self).__init__()

        self.layer1 = nn.Sequential(
            nn.Conv2d(1, 16, kernel_size=5, stride=1, padding=2),
            nn.LayerNorm([16] + input_shape), nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2))

        input_shape = [input_shape[0] // 2, input_shape[1] // 2]
        self.layer2 = nn.Sequential(
            nn.Conv2d(16, 32, kernel_size=5, stride=1, padding=2),
            nn.LayerNorm([32] + input_shape), nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2))

        self.fc = nn.Linear(7 * 7 * 32, num_classes)
        self.criterion = nn.CrossEntropyLoss()

    def forward(self, x, labels):
        out = self.layer1(x)
        out = self.layer2(out)
        out = out.reshape(out.size(0), -1)
        out = self.fc(out)
        loss = self.criterion(out, labels)
        return loss, out


def add_args(parser: argparse.ArgumentParser):
    parser.add_argument('--lr',
                        type=float,
                        default=0.001,
                        help="Learning rate for training")
    parser.add_argument('--momentum',
                        type=float,
                        default=0.0,
                        help="Momentum value for training")
    parser.add_argument('--weight-decay',
                        type=float,
                        default=1e-4,
                        help="Weight decay for training")
    parser.add_argument('-e', '--num-epochs', type=int, default=1)
    parser.add_argument('--num-classes', type=int, default=10)
    parser.add_argument(
        '--acc-test',
        action='store_true',
        help='Option for accuracy guard test in CH regression.')


def add_run_args(parser: argparse.ArgumentParser):
    parser.add_argument('--data-folder',
                        type=str,
                        default='mnist_data',
                        help="The folder to download the MNIST dataset to.")


def prepare_dataloader(args: argparse.Namespace
                       ) -> Tuple[torch.utils.data.DataLoader]:
    # Train the model on RDU using the MNIST dataset (images and labels)
    train_dataset = torchvision.datasets.MNIST(root=f'{args.data_folder}',
                                               train=True,
                                               transform=transforms.ToTensor(),
                                               download=True)
    test_dataset = torchvision.datasets.MNIST(root=f'{args.data_folder}',
                                              train=False,
                                              transform=transforms.ToTensor())

    # Data loader (input pipeline)
    train_loader = torch.utils.data.DataLoader(dataset=train_dataset,
                                               batch_size=args.batch_size,
                                               shuffle=True,
                                               drop_last=True)
    test_loader = torch.utils.data.DataLoader(dataset=test_dataset,
                                              batch_size=args.batch_size,
                                              shuffle=False,
                                              drop_last=True)
    return train_loader, test_loader


def train(args: argparse.Namespace, model: nn.Module,
          optimizer: samba.optim.SGD) -> None:
    train_loader, test_loader = prepare_dataloader(args)
    # Train the model
    total_step = len(train_loader)
    hyperparam_dict = {
        "lr": args.lr,
        "momentum": args.momentum,
        "weight_decay": args.weight_decay
    }
    for epoch in range(args.num_epochs):
        for i, (images, labels) in enumerate(train_loader):
            sn_images = samba.from_torch(images, name='image', batch_dim=0)
            sn_labels = samba.from_torch(labels, name='label', batch_dim=0)

            loss, outputs = samba.session.run(
                input_tensors=[sn_images, sn_labels],
                output_tensors=model.output_tensors,
                hyperparam_dict=hyperparam_dict,
                data_parallel=args.data_parallel,
                reduce_on_rdu=args.reduce_on_rdu)

            loss, outputs = samba.to_torch(loss), samba.to_torch(outputs)

            if (i + 1) % 10000 == 0:
                print('Epoch [{}/{}], Step [{}/{}], Loss: {:.4f}'.format(
                    epoch + 1, args.num_epochs, i + 1, total_step,
                    avg_loss / (i + 1)))

        model.eval()
        with torch.no_grad():
            correct = 0
            total = 0
            total_loss = 0
            for images, labels in test_loader:
                sn_images = samba.from_torch(images, name='image', batch_dim=0)
                sn_labels = samba.from_torch(labels, name='label', batch_dim=0)

                loss, outputs = samba.session.run(
                    input_tensors=[sn_images, sn_labels],
                    section_types=["fwd"],
                    output_tensors=model.output_tensors,
                    hyperparam_dict=hyperparam_dict,
                    data_parallel=args.data_parallel,
                    reduce_on_rdu=args.reduce_on_rdu)
                loss, outputs = samba.to_torch(loss), samba.to_torch(outputs)
                total_loss += loss.mean()
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum()

            test_acc = 100.0 * correct / total
            print(
                'Test Accuracy: {:.2f}'.format(test_acc),
                ' Loss: {:.4f}'.format(total_loss.item() / (len(test_loader))))

        if args.acc_test:
            assert args.num_epochs == 1, "Accuracy test only supported for 1 epoch"
            assert test_acc > 91.0 and test_acc < 92.0, "Test accuracy not within specified bounds."


def main(argv):
    utils.set_seed(256)
    args = parse_app_args(argv=argv,
                          common_parser_fn=add_args,
                          run_parser_fn=add_run_args)

    #rcwipt = samba.randn(args.batch_size, 28, 28, 1, name='image', batch_dim=0).bfloat16().float()
    ipt = samba.randn(args.batch_size, 1, 28, 28, name='image', batch_dim=0)
    tgt = samba.randint(args.num_classes, (args.batch_size, ),
                        name='label',
                        batch_dim=0)
    model = ConvNet(args.num_classes)

    samba.from_torch_(model)

    inputs = (ipt, tgt)
    # Instantiate a optimizer.
    if args.inference:
        optimizer = None
    else:
        optimizer = samba.optim.SGD(model.parameters(),
                                    lr=args.lr,
                                    momentum=args.momentum,
                                    weight_decay=args.weight_decay)

    if args.command == "compile":
        # Run model analysis and compile, this step will produce a PEF.
        samba.session.compile(model,
                              inputs,
                              optimizer,
                              name='cnn_mnist_torch',
                              squeeze_bs_dim=True,
                              app_dir=utils.get_file_dir(__file__),
                              config_dict=vars(args))
    elif args.command == "run":
        utils.trace_graph(model,
                          inputs,
                          optimizer,
                          pef=args.pef,
                          mapping=args.mapping)
        train(args, model, optimizer)


if __name__ == '__main__':
    main(sys.argv[1:])
